/*
  Mini bibliothque de fonction de transformation optique pour SDL
 
  Copyright (C) 2006 Humbert Florent http://www.developpez.net/forums/member.php?u=103584

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <math.h>

#include "SDL_transoptique.h"

/*
 *
 * Objet Couleur
 *
 */


typedef struct CouleurRGB
{
 Uint8 r;
 Uint8 g;
 Uint8 b;      
} Couleur;


/* max et min de a et b*/
static float maxi(float a, float b);
static float mini(float a, float b);

/* Retourne la couleur noir correspondant  la structure*/
static Couleur couleurNoir();


/*Creer une couleur RGB*/
static Couleur couleurCreer(Uint8, Uint8, Uint8);

/* retourne la couleur multiplie par un flottant*/
static Couleur couleurMult(Couleur couleur, float x);

/*Somme de deux couleurs*/
static Couleur couleurSomme(Couleur, Couleur);

/*inverse une couleur*/
static Couleur couleurInverser(Couleur);


/*
 *
 * Fonction de base sur les SDL_Surface
 *
 */

/*Lit la couleur au format Uint32 d'une surface*/
static Uint32 SDL_LireCouleurAux(SDL_Surface* surface, int x, int y);

/*Lit une couleur au format Couleur d'une surface*/
static Couleur SDL_LireCouleur(SDL_Surface * surface, int x, int y);

/*Permet d'crire sur une surface avec une couleur Uint32*/
static void SDL_EcrireCouleurAux(SDL_Surface* surface, int x, int y, Uint32 pixel);

/*Permet de lire sur une surface avec une couleur Couleur*/
static void SDL_EcrireCouleur(SDL_Surface* surface, int x, int y, Couleur couleur);


/*Dtermine la couleur d'un pixel  la position x, y d'une surface soumis  la convolution d'une matrice*/
static Couleur couleurConvolutionNM(SDL_Surface* surface, int x, int y, Matrice* matrice);







/*
 *
 * Dfinition des oprations sur les matrices
 *
 */


int matriceInitialiser(Matrice* matrice, int nbligne, int nbcolonne)
{
 float* m;
 int i;
 if(matrice==NULL)
  return EXIT_FAILURE;
 
 m = (float*) malloc(nbligne * nbcolonne * sizeof(float));
 
 if (m==NULL)
  return EXIT_FAILURE;
  
 matrice->m = m;
 matrice->nblignes = nbligne;
 matrice->nbcolonnes = nbcolonne;
 for(i=0;i<(nbligne*nbcolonne);i++)
  m[i] = 0.0F;
 
 return EXIT_SUCCESS;         
}

int matriceDetruire(Matrice* matrice)
{
 if (matrice->m ==NULL)
   return EXIT_FAILURE;

 free(matrice->m);
  
 return EXIT_SUCCESS;   
}

int matriceGetNbLignes(Matrice* matrice)
{
  return matrice->nblignes;
}

int matriceGetNbColonnes(Matrice* matrice)
{
   return matrice->nbcolonnes;      
}

void inline matriceEcrire(Matrice* m, int ligne, int colonne, float element)
{
 if (ligne>0 && colonne>0 && ligne<=matriceGetNbLignes(m) && colonne<=matriceGetNbColonnes(m))
   m->m[(ligne-1) * matriceGetNbColonnes(m) + (colonne-1)] =element ;    
}

float inline matriceLire(Matrice* m, int ligne, int colonne)
{
 return m->m[(ligne-1) * matriceGetNbColonnes(m) + (colonne-1)];     
}

float inline matriceCalculerSomme(Matrice* m)
{
 int i,j;
 float res = 0.0F;
 
 for(i=0;i<m->nblignes*m->nbcolonnes;i++)
   res += m->m[i];

 return res;      
}

void matriceDiviser(Matrice* m, float diviseur)
{
 int i,j;
 
 for(i=0;i<m->nblignes*m->nbcolonnes;i++)
    m->m[i] = m->m[i]/diviseur;
}





/*
 *
 * Dfinition des oprations sur le type Couleur
 *
 */

float inline mini(float a, float b)
{
 if (a<b)
  return a;
 else
  return b;    
}

float inline maxi(float a, float b)
{
 if (a<b) 
  return b ;
    else 
 return a;      
}


Couleur inline couleurNoir()
{
 Couleur couleur;
 couleur.r = 0;
 couleur.g = 0;
 couleur.b = 0;
 return couleur;        
}

Couleur inline couleurCreer(Uint8 r, Uint8 g, Uint8 b)
{
 Couleur dest;
 dest.r = r;
 dest.g = g;
 dest.b = b;

 return dest;        
} 
 
Couleur inline couleurMult(Couleur couleur, float x)
{
 Couleur dest;
 dest.r = (Uint8) ((float) ((float) couleur.r) * x);
 dest.g = (Uint8) ((float) ((float) couleur.g) * x);
 dest.b = (Uint8) ((float) ((float) couleur.b) * x);
 return dest;         
          
} 

Couleur inline couleurInverser(Couleur c)
{
 Couleur dest;
 dest.r = ((Uint8) 255) -c.r;
 dest.g = ((Uint8) 255) -c.g;
 dest.b = ((Uint8) 255) -c.b;
 
 return dest;
        
}
        

Couleur inline couleurSomme(Couleur c1, Couleur c2)
{
 Couleur dest;
 dest.r = (Uint8) mini(255.0f, (float) c1.r + c2.r);
 dest.g = (Uint8) mini(255.0f, (float) c1.g + c2.g);
 dest.b = (Uint8) mini(255.0f, (float) c1.b + c2.b);
 
 return dest;        
}













/*
 *
 * Dfinition des oprations sur le type SDL_Surface
 *
 */

Uint32 inline SDL_LireCouleurAux(SDL_Surface* surface, int x, int y)
{
  int bpp = surface->format->BytesPerPixel;

  Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

  switch(bpp)
  {
             case 1:
                  return *p;
             case 2:
                  return *(Uint16 *)p;
             case 3:
                 if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
                     return p[0] << 16 | p[1] << 8 | p[2];
                 else
                     return p[0] | p[1] << 8 | p[2] << 16;
             case 4:
                  return *(Uint32 *)p;
             default:
                  return 0;
  }        
}



Couleur inline SDL_LireCouleur(SDL_Surface * surface, int x, int y)
{
 Couleur couleur;
 if (x<0 | y<0 | x>=(surface->w) | y>=(surface->h))
  return couleurNoir();
 
 SDL_GetRGB(SDL_LireCouleurAux(surface, x, y), surface->format, 
                                      (Uint8*)   &couleur.r, (Uint8*) &couleur.g, (Uint8*) &couleur.b);
 return couleur;
          
          
}


void inline SDL_EcrireCouleurAux(SDL_Surface* surface, int x, int y, Uint32 pixel)
{
     int bpp = surface->format->BytesPerPixel; 
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp; 
  
  
    switch(bpp) { 
    case 1: 
        *p = pixel; 
        break; 
  
    case 2: 
        *(Uint16 *)p = pixel; 
        break; 
  
    case 3: 
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) { 
            p[0] = (pixel >> 16) & 0xff; 
            p[1] = (pixel >> 8) & 0xff; 
            p[2] = pixel & 0xff; 
        } else { 
            p[0] = pixel & 0xff; 
            p[1] = (pixel >> 8) & 0xff; 
            p[2] = (pixel >> 16) & 0xff; 
        } 
        break; 
  
    case 4: 
        *(Uint32 *)p = pixel; 
        break; 
    }    
}

void inline SDL_EcrireCouleur(SDL_Surface* surface, int x, int y, Couleur couleur)
{
 SDL_EcrireCouleurAux(surface, x, y, 
    SDL_MapRGB(surface->format, couleur.r, couleur.g, couleur.b));
}











/*
 *
 * Convolution NM
 *
 *
 *
 *
 *
 */

Couleur inline couleurConvolutionNM(SDL_Surface* surface, int x, int y, Matrice* matrice)
{
 float r = 0;
 float g = 0;
 float b = 0;
 int i,j;
 Couleur coulGet;  /*aucun rapport avec la marque de dentifrice*/
 
 int m_lignes = matriceGetNbLignes(matrice);
 int m_colonnes = matriceGetNbColonnes(matrice);
 int m_milieu_lignes = (m_lignes-1) /2+1;
 int m_milieu_colonnes = (m_colonnes-1)/2+1;

 for (i=(m_milieu_lignes-m_lignes);i<=(m_lignes-m_milieu_lignes);i++)
  for (j=(m_milieu_colonnes-m_colonnes);j<=(m_lignes-m_milieu_colonnes);j++)
   {
    coulGet = SDL_LireCouleur(surface, x+i, y+j);
     r += ((float) coulGet.r) * matriceLire(matrice, i+m_milieu_lignes,j+m_milieu_colonnes);  
     g += ((float) coulGet.g) * matriceLire(matrice, i+m_milieu_lignes,j+m_milieu_colonnes);  
     b += ((float) coulGet.b) * matriceLire(matrice, i+m_milieu_lignes,j+m_milieu_colonnes);  
                   
   }           
 coulGet.r = (Uint8) mini(maxi(r,0.0f),255.0f);
 coulGet.g = (Uint8) mini(maxi(g,0.0f),255.0f);
 coulGet.b = (Uint8) mini(maxi(b,0.0f),255.0f);
 
 return coulGet;
}




SDL_Surface* convolveNM(SDL_Surface* origine, Matrice* matrice)
{
  SDL_Surface* destination;
  int i,j;            
  int hauteur = origine->h;
  int largeur = origine->w;
 
  Couleur cdest;
  if (origine==NULL)
   return NULL;
   
  destination = SDL_CreateRGBSurface(SDL_HWSURFACE, origine->w, origine->h, origine->format->BitsPerPixel,
			origine->format->Rmask, origine->format->Gmask, origine->format->Bmask, origine->format->Amask);

  if (destination==NULL)
   return NULL;
 
  for(i=0; i<(largeur); i++)
   for (j=0; j<(hauteur); j++)
   {
    
       
    cdest = couleurConvolutionNM(origine, i, j,matrice);
    SDL_EcrireCouleur(destination, i, j, cdest);

  }
  
 return destination;          
}






/*
 * Dtection de bords avec l'oprateur Laplacien
 */

SDL_Surface* SDL_DetectionsBords(SDL_Surface * origine)
{
  Matrice m;
  SDL_Surface* surf;
  matriceInitialiser(&m, 3,3);

  matriceEcrire(&m, 1, 1, 0);
  matriceEcrire(&m, 1, 2, 1);
  matriceEcrire(&m, 1, 3, 0);
  matriceEcrire(&m, 2, 1, 1);
  matriceEcrire(&m, 2, 2, -4);
  matriceEcrire(&m, 2, 3, 1);
  matriceEcrire(&m, 3, 1, 0);
  matriceEcrire(&m, 3, 2, 1);
  matriceEcrire(&m, 3, 3, 0);

 
  surf = convolveNM(origine, &m);    
  matriceDetruire(&m);     

  return surf;        
}



/*
 * Gaufrage
 */
SDL_Surface* SDL_Gaufrage(SDL_Surface* origine)
{
   Couleur ori;
   int i,j;
   SDL_Surface* dest;
   Matrice m;
  
  
   matriceInitialiser(&m, 3,3); 

   matriceEcrire(&m, 1, 1, -2);
   matriceEcrire(&m, 1, 2, -1);
   matriceEcrire(&m, 1, 3, 0);
   matriceEcrire(&m, 2, 1, -1);
   matriceEcrire(&m, 2, 2, 0);
   matriceEcrire(&m, 2, 3, 1);
   matriceEcrire(&m, 3, 1, 0);
   matriceEcrire(&m, 3, 2, 1);
   matriceEcrire(&m, 3, 3, 2);
   
   dest = convolveNM(origine, &m); 
 
 
   if (dest==NULL)
     return NULL;
      
   for(j=0;j< dest->h;j++)
    for(i=0;i< dest->w;i++)
    {
      ori = SDL_LireCouleur(dest, i, j);
      ori = couleurSomme(ori, couleurCreer((Uint8) 128,(Uint8) 128,(Uint8) 128));
      SDL_EcrireCouleur(dest, i,j, ori);            
    }       
 matriceDetruire(&m);
 return dest;
}

/*
 * Repoussage
 */
SDL_Surface* SDL_Repoussage(SDL_Surface* origine)
{
  Matrice m;
  SDL_Surface* surf;
    
  matriceInitialiser(&m, 3,3); 
  
  matriceEcrire(&m, 1, 1, -2);
  matriceEcrire(&m, 1, 2, -1);
  matriceEcrire(&m, 1, 3, 0);
  matriceEcrire(&m, 2, 1, -1);
  matriceEcrire(&m, 2, 2, 1);
  matriceEcrire(&m, 2, 3, 1);
  matriceEcrire(&m, 3, 1, 0);
  matriceEcrire(&m, 3, 2, 1);
  matriceEcrire(&m, 3, 3, 2);


 
  surf = convolveNM(origine, &m);    
  matriceDetruire(&m);    

  return surf;        
}

/*
SDL_Surface* SDL_PasseBas(SDL_Surface* origine)
{
  Matrice m;
  SDL_Surface* surf;
    
  matriceInitialiser(&m, 3,3); 
  
  matriceEcrire(&m, 1, 1, 0.0625F);
  matriceEcrire(&m, 1, 2, 0.125F);
  matriceEcrire(&m, 1, 3, 0.0625F);
  matriceEcrire(&m, 2, 1, 0.125F);
  matriceEcrire(&m, 2, 2, 0.25F);
  matriceEcrire(&m, 2, 3, 0.125F);
  matriceEcrire(&m, 3, 1, 0.0625F);
  matriceEcrire(&m, 3, 2, 0.125F);
  matriceEcrire(&m, 3, 3, 0.0625F);


 
  surf = convolveNM(origine, &m);    
  matriceDetruire(&m);    

  return surf;        
}*/

/*
 * Amlioration des phrases
 */
SDL_Surface* SDL_AmeliorationBords(SDL_Surface* origine)
{
  Matrice m;
  SDL_Surface* surf;
    
  matriceInitialiser(&m, 3,3); 
  
  matriceEcrire(&m, 1, 1, 0);
  matriceEcrire(&m, 1, 2, 0);
  matriceEcrire(&m, 1, 3, 0);
  matriceEcrire(&m, 2, 1, -1);
  matriceEcrire(&m, 2, 2, 2);
  matriceEcrire(&m, 2, 3, 0);
  matriceEcrire(&m, 3, 1, 0);
  matriceEcrire(&m, 3, 2, 0);
  matriceEcrire(&m, 3, 3, 0);
 
  surf = convolveNM(origine, &m);    
  matriceDetruire(&m);    
  
  return surf;       
}


/*
 * Augmentation du constraste
 */
SDL_Surface* SDL_AugmenterConstraste(SDL_Surface* origine)
{
  Matrice m;
  SDL_Surface* surf;
    
  matriceInitialiser(&m, 3,3); 
  
  matriceEcrire(&m, 1, 1, 0);
  matriceEcrire(&m, 1, 2, -1);
  matriceEcrire(&m, 1, 3, 0);
  matriceEcrire(&m, 2, 1, -1);
  matriceEcrire(&m, 2, 2, 5);
  matriceEcrire(&m, 2, 3, -1);
  matriceEcrire(&m, 3, 1, 0);
  matriceEcrire(&m, 3, 2, -1);
  matriceEcrire(&m, 3, 3, 0);
 
  surf = convolveNM(origine, &m);    
  matriceDetruire(&m);  
  
  return surf;     
}


/*
 * Flou de base (appel aussi lissage)
 */
SDL_Surface* SDL_Flou(SDL_Surface* origine)
{
  Matrice m;
  SDL_Surface* surf;

  matriceInitialiser(&m, 3,3); 
  
  matriceEcrire(&m, 1, 1, 0.111111);
  matriceEcrire(&m, 1, 2, 0.111111);
  matriceEcrire(&m, 1, 3, 0.111111);
  matriceEcrire(&m, 2, 1, 0.111111);
  matriceEcrire(&m, 2, 2, 0.111111);
  matriceEcrire(&m, 2, 3, 0.111111);
  matriceEcrire(&m, 3, 1, 0.111111);
  matriceEcrire(&m, 3, 2, 0.111111);
  matriceEcrire(&m, 3, 3, 0.111111);  
 
  surf = convolveNM(origine, &m);    
  matriceDetruire(&m);      
  
  return surf;    
}


/*
 * inverse les couleurs d'une surface
 */
SDL_Surface* SDL_Inverser(SDL_Surface* origine)
{
 int i,j;
 Couleur cdest;
 SDL_Surface* destination;

 int hauteur = origine->h;
 int largeur = origine->w;
 
 if (origine==NULL)
   return NULL;
   
  destination = SDL_CreateRGBSurface(SDL_HWSURFACE, origine->w, origine->h, origine->format->BitsPerPixel,
			origine->format->Rmask, origine->format->Gmask, origine->format->Bmask, origine->format->Amask);

  if (destination==NULL)
   return NULL;

 for(i=0;i<largeur;i++)
  for(j=0;j<hauteur;j++)
  {
    cdest = SDL_LireCouleur(origine, i,j);
    cdest = couleurInverser(cdest);
    SDL_EcrireCouleur(destination, i, j, cdest);                    
  }   
 
 return destination;            
             
}


float carre(float x)
{
  return (x*x);     
}

SDL_Surface* SDL_FlouGaussien(SDL_Surface* origine, float sigma, unsigned int taille_noyau)
{
 Matrice noyau;
 SDL_Surface* dest;
 int i,j; 
 
 int m_lignes;
 int m_colonnes;
 int m_milieu_lignes;
 int m_milieu_colonnes;

 
 if((taille_noyau %2) == 0)
  taille_noyau++;
  
   
 if(sigma==0)
 {
  fprintf(stderr, " sigma nulle, division par 0");
  return NULL;
 }
 
 /*gnration du noyau*/
 if(matriceInitialiser(&noyau, taille_noyau, taille_noyau)==EXIT_FAILURE)
 {
   fprintf(stderr, " * matrice : manque mmoire");
   return NULL;
 }

 m_lignes = matriceGetNbLignes(&noyau);
 m_colonnes = matriceGetNbColonnes(&noyau);
 m_milieu_lignes = (m_lignes-1) /2+1;
 m_milieu_colonnes = (m_colonnes-1)/2+1;


 for(i=1;i<=m_lignes;i++)
  for(j=1;j<=m_colonnes;j++)
  {
    matriceEcrire(&noyau, i, j, 
        1/(2*M_PI*carre(sigma)) * exp(-(carre(i-m_milieu_lignes)+carre(j-m_milieu_colonnes))/(2*carre(sigma)))
  //-1/(M_PI * carre(carre(sigma)))* (1- (carre(i-m_milieu_lignes)+carre(j-m_milieu_colonnes))/
  //           (2*carre(sigma))) * exp(-(carre(i-m_milieu_lignes)+carre(j-m_milieu_colonnes))/(2*carre(sigma)))
  
  // 1/(carre(i-m_milieu_lignes)+carre(j-m_milieu_colonnes)+1)
        );
  } 
 
 //matriceDiviser(&noyau, matriceCalculerSomme(&noyau));
  
 dest =  convolveNM(origine, &noyau);            
 matriceDetruire(&noyau);
 
 return dest;            
             
}

/*
SDL_Surface* SDL_ConvolutionFonction(SDL_Surface* origine, void (* fonction))
{

 return NULL;             
}*/


/*
 * Seuillage binaire
 */
SDL_Surface* SDL_SeuillageBinaire(SDL_Surface* origine, Uint8 r, Uint8 g, Uint8 b)
{
 int i,j;
 Couleur cdest;
 SDL_Surface* destination;

 int hauteur = origine->h;
 int largeur = origine->w;
 
 if (origine==NULL)
   return NULL;
   
  destination = SDL_CreateRGBSurface(SDL_HWSURFACE, origine->w, origine->h, origine->format->BitsPerPixel,
			origine->format->Rmask, origine->format->Gmask, origine->format->Bmask, origine->format->Amask);

  if (destination==NULL)
   return NULL;

 for(i=0;i<largeur;i++)
  for(j=0;j<hauteur;j++)
  {
    cdest = SDL_LireCouleur(origine, i, j);
    if(r < cdest.r)
      cdest.r = (Uint8) 255;
    else
      cdest.r = (Uint8) 0;
      
    if(g < cdest.g)
      cdest.g = (Uint8) 255;
    else
      cdest.g = (Uint8) 0;

    if(b < cdest.b)
      cdest.b = (Uint8) 255;
    else
      cdest.b = (Uint8) 0;
                            
    SDL_EcrireCouleur(destination, i, j, cdest);                    
  }

 return destination;

}


/*
 * Griser une surface
 */
SDL_Surface* SDL_Griser(SDL_Surface* origine)
{
 int i,j;
 Couleur cdest;
 SDL_Surface* destination;

 int hauteur = origine->h;
 int largeur = origine->w;
 
 if (origine==NULL)
   return NULL;
   
  destination = SDL_CreateRGBSurface(SDL_HWSURFACE, origine->w, origine->h, origine->format->BitsPerPixel,
			origine->format->Rmask, origine->format->Gmask, origine->format->Bmask, origine->format->Amask);

  if (destination==NULL)
   return NULL;

 for(i=0;i<largeur;i++)
  for(j=0;j<hauteur;j++)
  {
    cdest = SDL_LireCouleur(origine, i, j);
    cdest.r = cdest.g = cdest.b = (Uint8) ((((float) (float) cdest.r + (float) cdest.g + (float) cdest.b))/3);                        
    SDL_EcrireCouleur(destination, i, j, cdest);                    
  }

 return destination;

}





















int main(int argc, char *argv[])
{
  SDL_Surface* fond;
  SDL_Surface* origine;
  SDL_Surface* destination;
  SDL_Surface* dest2;
  SDL_Event evenement;
  int done = 0;
  
   SDL_Init(SDL_INIT_VIDEO);	
   fond = SDL_SetVideoMode(1000,800,  32,SDL_HWSURFACE|SDL_DOUBLEBUF);

 if (fond==NULL)
   return EXIT_FAILURE;
 
 origine = SDL_LoadBMP("03.bmp");
 
 if (origine==NULL)
   {
   fprintf(stderr,"Erreur dans le chargement de l'image\n");
   }
 SDL_BlitSurface(origine, NULL, fond, NULL);
 SDL_Flip(fond);
 
 /*application de la dtection de bords*/
 destination = SDL_DetectionsBords(origine);
// destination = SDL_SeuillageBinaire(origine, 100, 100, 100) ;
// destination = SDL_FlouGaussien(origine,0.05,15);
 //destination = SDL_Inverser(destination);
 //SDL_FreeSurface(destination);
 //destination = SDL_AugmenterConstraste(dest2);
 //origine = SDL_AugmenterConstraste(destination);
 //SDL_SaveBMP(destination, "ambords.bmp");
  
 SDL_BlitSurface(destination, NULL, fond, NULL);
 SDL_Flip(fond);
 
 while(!done)
    {
        //Traiter les evenements
	while(SDL_PollEvent(&evenement))
	{
		switch(evenement.type)
		{
		case SDL_QUIT:
			done=1;
			break;
   	        case SDL_KEYUP:
			if(evenement.key.keysym.sym==SDLK_q)
				done=1;
			if(evenement.key.keysym.sym==SDLK_LEFT)
            { 
             SDL_BlitSurface(destination, NULL, fond, NULL);
             SDL_Flip(fond);
            }
            if(evenement.key.keysym.sym==SDLK_RIGHT)
            { 
             SDL_BlitSurface(origine, NULL, fond, NULL);
             SDL_Flip(fond);
            }  
			break;
		default:
			break;
		}
	}
  }
  
  
  SDL_FreeSurface(destination);
  SDL_FreeSurface(origine);
  SDL_Quit();
  (void) argc;
  (void) argv;
  return EXIT_SUCCESS;
}
