/*
  Mini bibliothque de fonction de transformation optique pour SDL
 
  Copyright (C) 2006 Humbert Florent http://www.developpez.net/forums/member.php?u=103584

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#ifndef _SDL_TRANSFORMATION_OPTIQUE_H
#define _SDL_TRANSFORMATION_OPTIQUE_H

#include <SDL/SDL.h>


/* */
/*  Auteur : Humbert Florent*/


/* Attention, chacune de ces fonctions alloue de l'espace pour l'image modifie,
 * n'oubliez pas de faire un SDL_FreeSurface() sur celle-ci
 */

/*Objet Matrice*/
typedef struct Matrice_
{
 float * m;
 int nblignes;
 int nbcolonnes;               
} Matrice;

/*Initialise une matrice selon son nombre de ligne et nombre de colonne*/
int matriceInitialiser(Matrice*, int nbligne, int nbcolonne);

/*Ecrit un float  une certaine ligne et certain colonne, la numrotation commence 1*/
void matriceEcrire(Matrice*, int ligne, int colonne, float element);

/*Lit dans une matrice  une certaine ligne et colonne*/
float matriceLire(Matrice*, int ligne, int colonne);

/*Retourne le nombre de ligne de la matrice*/
int matriceGetNbLignes(Matrice*);

/*Retourne le nombre de colonne de la matrice*/
int matriceGetNbColonnes(Matrice*);

/*Vide la matrice*/
static int matriceDetruire(Matrice* matrice);

/* Calcul la somme de tous les coefficients d'une matrice*/
float matriceCalculerSomme(Matrice* m);

/*Divise tous les coefficients de la matrice par un diviseur*/
void matriceDiviser(Matrice* m, float diviseur);




/*applique une matrice de convolution N M  une surface et retourne la surface correspondante*/
SDL_Surface* convolveNM(SDL_Surface* origine, Matrice* matrice);

/*Ralise une transformation de dtetction de bords sur une surface*/
SDL_Surface* SDL_DetectionsBords(SDL_Surface * origine);

/*Ralise un gauffrage sur une surface*/
SDL_Surface* SDL_Gaufrage(SDL_Surface* origine);

/*Ralise une transformation de repoussage*/
SDL_Surface* SDL_Repoussage(SDL_Surface* origine);

/*Amliore les bords*/
SDL_Surface* SDL_AmeliorationBords(SDL_Surface* origine);

/*Augmente le contraste*/
SDL_Surface* SDL_AugmenterConstraste(SDL_Surface* origine);

/*Inverse une image*/
SDL_Surface* SDL_Inverser(SDL_Surface* origine);

/*Applique un flou  une image*/
SDL_Surface* SDL_Flou(SDL_Surface* origine);

/*Applique un flou Gaussien  une image.
 La grandeur de sigma doit tre proportionnel  la grosseur du noyau de convolution
 Par exemple: sigma =2, noyau=9 ou sigma=4 et noyau=15*/
SDL_Surface* SDL_FlouGaussien(SDL_Surface* origine, float sigma, unsigned int taille_noyau);

/*effectue un seuillage binaire*/
SDL_Surface* SDL_SeuillageBinaire(SDL_Surface* origine, Uint8 r, Uint8 g, Uint8 b);

/* transforme en niveau de gris une surface*/
SDL_Surface* SDL_Griser(SDL_Surface* origine);


#endif
